﻿using ILNumerics;
using ILNumerics.Drawing;
using ILNumerics.Drawing.Plotting;
using ILNumerics.IO.HDF5;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Imaging;
using static ILNumerics.Globals;
using static ILNumerics.ILMath;

const string h5Name = "values.h5";  

if (!File.Exists(h5Name)) {
     
    //ILN(enabled = false)
    Settings.ArrayStyle = ArrayStyles.numpy;
    Array<uint> A = counter<uint>(1507, 200, 507, 10, 5, 17);    // creates [507, 10, 5, 17] array. 
    Array<uint> B = counter<uint>(1507, 1, 1, 1, 5, 17);  // creates [1, 1, 5, 17] array.  

    const uint m0 = 1142251548;      // some variables to be used in our expression
    const int shift = 17;            // if variables are constants, they should be declared as such!
                                     // ILNumerics Accelerator will make use of it. 
    DateTime startTime;

    Array<uint> result = 1;          // A local variable will be used for testing correctness.
    List<Tuple<TimeSpan, double>> ongoing = new List<Tuple<TimeSpan, double>>(); // collects samples

    using var file = new H5File("values.h5");

    for (startTime = DateTime.Now; (DateTime.Now - startTime).Seconds < 10;) {

        var start = Stopwatch.StartNew();

        for (int i = 0; i < 1000; ++i) {
            result.a = sum((m0 & (A << shift)) | (~m0 & B), dim: 1);    // Original expression
        }

        var t = Tuple.Create(DateTime.Now - startTime, (double)start.ElapsedTicks / Stopwatch.Frequency); // gives ms
        ongoing.Add(t);
        System.Console.WriteLine($"Avg time msec (original): {t:F8} ms. result: " + result.GetValue(0));
    }
    fileIt(ongoing, file, "ILN_noAcc");

    //ILN(enabled = true)
    Array<uint> result2 = 1; ongoing.Clear();
    for (startTime = DateTime.Now; (DateTime.Now - startTime).Seconds < 10;) {

        var start = Stopwatch.StartNew();

        for (int i = 0; i < 1000; ++i) {
            result2.a = sum((m0 & (A << shift)) | (~m0 & B), dim: 1);    // expression will be optimized
        }
        result2.Finish();            // <-- only required for timing measurements!! This actually slows down
                                     // computations and should not be used in production code, unless required! 

        var t = Tuple.Create(DateTime.Now - startTime, (double)start.ElapsedTicks / Stopwatch.Frequency); // gives ms
        ongoing.Add(t);
        System.Console.WriteLine($"Avg time msec (optimized): {t:F8} ms. result: " + result2.GetValue(0));
    }
    fileIt(ongoing, file, "ILN_Acc");

    Segment.Default.SpecializeFlags = ILNumerics.Core.Segments.SpecializeFlags.BSDsAll; // <- experimental feature
    ongoing.Clear();
    for (startTime = DateTime.Now; (DateTime.Now - startTime).Seconds < 10;) {

        var start = Stopwatch.StartNew();

        for (int i = 0; i < 1000; ++i) {
            result2.a = sum((m0 & (A << shift)) | (~m0 & B), dim: 1);    // expression will be optimized more aggressively
        }
        result2.Finish();

        var t = Tuple.Create(DateTime.Now - startTime, (double)start.ElapsedTicks / Stopwatch.Frequency); // gives ms
        ongoing.Add(t);
        System.Console.WriteLine($"Avg time msec (optimized++): {t:F8} ms. result: " + result2.GetValue(0));
    }
    fileIt(ongoing, file, "ILN_AccBSDAll");

    Console.WriteLine($"Results equal: {(allall(result2 == result) ? "true" : "false")}");
    Console.WriteLine("Segment invocations count: " + Segment.RunCount);
}
draw(); 

static void fileIt(List<Tuple<TimeSpan, double>> ongoing, H5File file, string dsname) {
    using var s = Scope.Enter(ArrayStyles.ILNumericsV4);
    Array<double> m = (Array<double>)ongoing.Select(i => i.Item1.TotalMilliseconds).ToArray();
    m[":",1] = (Array<double>)ongoing.Select(i => (double)i.Item2).ToArray();
    file.Add(new H5Dataset(dsname, m));
}

void draw() {
    using var h5file = new H5File("values.h5");
    Array<double> ILN_noAcc = h5file.Get<H5Dataset>("ILN_noAcc").Get<double>(); 
    Array<double> ILN_Acc = h5file.Get<H5Dataset>("ILN_Acc").Get<double>();
    Array<double> ILN_AccBSDAll = h5file.Get<H5Dataset>("ILN_AccBSDAll").Get<double>();
    Array<double> Numpy = csvread<double>(File.ReadAllText("..\\..\\..\\..\\numpy\\values.py"));
    Array<double> Numpy_numba = csvread<double>(File.ReadAllText("..\\..\\..\\..\\numpy\\values_numba.py"));
    Array<double> fortran = csvread<double>(File.ReadAllText(@"..\..\..\..\FORTRAN\values_fortran.txt"), elementSeparator: "      ");

    ILNumerics.Drawing.Label.DefaultFont = new Font("Roboto", 14); 
    var scene = new Scene {
        new PlotCube(twoDMode: true) {
            new LinePlot(tosingle(ILN_noAcc).T, tag: "ILNumerics, V7, no acceleration", lineWidth: 3, markerStyle: MarkerStyle.Square),
            new LinePlot(tosingle(ILN_Acc).T[full,r(0,2,end)], tag: "ILNumerics, V7, accelerator", lineWidth: 3, markerStyle: MarkerStyle.Square),
            new LinePlot(tosingle(ILN_AccBSDAll).T, tag: "ILNumerics, V7, accelerator:BSDAll", lineWidth: 3, markerStyle: MarkerStyle.None),
            new LinePlot(tosingle(Numpy).T, tag: "Numpy 2.0.2", lineWidth: 3, markerStyle: MarkerStyle.Dot),
            new LinePlot(tosingle(Numpy_numba).T, tag: "Numpy 2.0.2, numba 0.60.0, ", lineWidth: 3, markerStyle: MarkerStyle.Dot),
            new LinePlot(tosingle(fortran).T[full,r(0,2,end)], tag: "Fortran ifort 2025", lineWidth: 3, markerStyle: MarkerStyle.TriangleUp),
            new Legend() {
                Location = new PointF(.9f, .32f)
            }
        }
    };
    var pc = scene.First<PlotCube>();
    pc.ScaleModes.YAxisScale = AxisScale.Logarithmic;
    pc.Axes.XAxis.Label.Text = "Time since app started [s]";
    pc.Axes.YAxis.Label.Text = "Execution time per cycle [ms]"; 
    pc.Axes.XAxis.Ticks.Clear();
    pc.Axes.XAxis.Ticks.DefaultLabel.Anchor = new PointF(.5f, 0f);
    //pc.Axes.XAxis.Ticks.DefaultLabel.Font = new Font("Arial", 15f, FontStyle.Bold);
    pc.Axes.XAxis.Ticks.Add(0, "0");
    pc.Axes.XAxis.Ticks.Add(2000, "2");
    pc.Axes.XAxis.Ticks.Add(4000, "4");
    pc.Axes.XAxis.Ticks.Add(6000, "6");
    pc.Axes.XAxis.Ticks.Add(8000, "8");
    pc.Axes.XAxis.Ticks.Add(10000, "10");
    pc.Axes.XAxis.Ticks.Mode = TickMode.Manual;

    var title = pc.Add(new Title(
@"\bfSpeed comparison: running low level expres-
sions with ILNumerics, numpy and FORTRAN
\reset
Intel i7 12700K 3.6GHz 12 cores
64GB DDR3 RAM, Windows 11
data: [507, 10, 5, 17] uint32"));
    title.Location = new PointF(.14f, .40f);
    title.Anchor = new PointF(0, 0.5f); 
    title.Label.Font = new Font(title.Label.Font.FontFamily, 14f);

    scene.Configure();
    var gdi = new GDIDriver(1500, 1000);
    gdi.Scene = scene;
    gdi.Render();
    gdi.BackBuffer.Bitmap.Save("Getting_Started_I.jpg", ImageFormat.Jpeg);

    using var outStream = new FileStream("Getting_Started_I.svg", FileMode.Create); 
    new SVGDriver(outStream, 1500, 1000, scene: scene).Render(); 

} 